#!/usr/bin/env python

import dbus
import dbus.glib
import gobject
import sys

import pygst
pygst.require("0.10")
import gst

import farsight
import fs2

from util import *

from telepathy.client.channel import Channel
from telepathy.constants import (
    CONNECTION_HANDLE_TYPE_NONE, CONNECTION_HANDLE_TYPE_CONTACT,
    CONNECTION_STATUS_CONNECTED, CONNECTION_STATUS_DISCONNECTED,
    MEDIA_STREAM_STATE_CONNECTED
    )
from telepathy.interfaces import (
    CHANNEL_INTERFACE, CONN_INTERFACE,
    CONNECTION_INTERFACE_REQUESTS,
    CONNECTION_INTERFACE_CONTACT_CAPABILITIES,
    CLIENT)

from constants import *

class FsStream:
    def __init__ (self, participant, conference, session):
        self.participant = participant
        self.conference = conference
        self.session = session
        self.remote_codecs = []
        self.src_pad_hook = None

        #setup a dummy transmitter
        self.stream = session.add_participant (participant, "multicast")

    def change_stream (self, transmitter, transmitter_params = {}):
        self.conference.remove_stream (self.stream)
        self.stream = self.session.add_participant (self.participant,
            transmitter, transmitter_params)

        if (self.remote_codecs != []):
            self.stream.set_remote_codecs (self.remote_codecs)

        if self.src_pad_hook != None:
            self.stream.src_pad_hook = self.src_pad_hook

    def set_remote_codecs (self, fscodecs):
        if fscodecs == self.remote_codecs:
            return

        self.remote_codecs = fscodecs
        self.stream.set_remote_codecs (fscodecs)

    def add_remote_candidate (self, fscandidate):
        self.stream.add_remote_candidate (fscandidate)

    def set_direction (self, direction):
        self.stream.set_direction(direction)

    def set_src_pad_added_callback (self, func):
        self.stream.src_pad_hook = func
        self.src_pad_hook = func

    def destroy(self):
        self.conference.remove_stream(self.stream)

class CallStream:
    def __init__(self, bus, connection, obj, object_path, fsstream):
        self.conn = connection
        self.bus = bus
        self.obj = obj
        self.object_path = object_path
        self.endpoints = []
        self.fsstream = fsstream
        self.senders_changed_cb = None

        retrieved_server_info = self.obj.Get (CALL_STREAM_IFACE_MEDIA,
            "RetrievedServerInfo", dbus_interface=dbus.PROPERTIES_IFACE)

        if retrieved_server_info == False:
            print "Wait for server info"
            self.obj.connect_to_signal ("ServerInfoRetrieved",
                self.server_info_retrieved,
                dbus_interface = CALL_STREAM_IFACE_MEDIA)
        else:
            self.server_info_retrieved ()

    def server_info_retrieved (self):
        media_properties = self.obj.GetAll (CALL_STREAM_IFACE_MEDIA,
            dbus_interface=dbus.PROPERTIES_IFACE)

        endpoints = media_properties["Endpoints"]
        assert (len(endpoints) == 1)

        transport = media_properties["Transport"]
        stun_servers = media_properties["STUNServers"]
        relay_info = media_properties["RelayInfo"]

        relays = []
        for info in relay_info:
            relay = gst.Structure ("relay")
            for key, value in info.items ():
                if key == 'component' or key == 'port':
                    relay.set_value (key, value, 'uint')
                else:
                    relay.set_value (key, value)
            relays.append (relay)

        self.setup_stream (transport, stun_servers, relays)
        self.setup_endpoint (endpoints[0], transport)

    def setup_endpoint (self, object_path, transport):
        endpoint = self.bus.get_object (self.conn.service_name, object_path)
        self.endpoints.append(endpoint)
        endpoint.connect_to_signal ('RemoteCandidatesAdded',
            self.remote_candidates_added,
            dbus_interface = CALL_STREAM_ENDPOINT)

        properties = endpoint.GetAll (CALL_STREAM_ENDPOINT,
            dbus_interface=dbus.PROPERTIES_IFACE)

        self.remote_candidates_added (
            properties["RemoteCandidates"])

        self.fsstream.stream.componentStateChanged.addCallback (False,
            self.component_state_changed, endpoint)

        if transport == STREAM_TRANSPORT_RAW_UDP:
          endpoint.SetStreamState (MEDIA_STREAM_STATE_CONNECTED,
            dbus_interface=CALL_STREAM_ENDPOINT)

    def component_state_changed (self, endpoint, component, state):
        if component == 1 and state == farsight.STREAM_STATE_READY:
            endpoint.SetStreamState (
                MEDIA_STREAM_STATE_CONNECTED,
                dbus_interface=CALL_STREAM_ENDPOINT)

    def remote_candidates_added (self, candidates):
        TP2FS = {
            "Type": ("type", int),
            "Foundation": ("foundation", str),
            "Protocol": ("proto", int),
            "Priority": ("priority", int),
            "Username": ("username", str),
            "Password": ("password", str),
            "BaseIP": ("base_ip", str),
            "BasePort": ("base_port", int),
        }
        for c in candidates:
            fscandidate = farsight.Candidate()
            fscandidate.component_id = int(c[0])
            fscandidate.ip = str(c[1])
            fscandidate.port = int(c[2])
            info = c[3]
            for (key, value) in info.iteritems():
                if TP2FS.has_key (key):
                    (attr, conv) = TP2FS[key]
                    setattr (fscandidate, attr, conv(value))
            self.fsstream.add_remote_candidate (fscandidate)

    def setup_stream(self, transport, stun_servers, relay_info):
        transmitters = {
            STREAM_TRANSPORT_GTALK_P2P: ("nice", { 'compatibility-mode':  1 }),
            STREAM_TRANSPORT_RAW_UDP: ("rawudp", {} ),
            STREAM_TRANSPORT_ICE_UDP: ("nice", {})
        }
        # FIXME set up controlling mode
        (transmitter, transmitter_params) = transmitters[transport]

        if len(stun_servers) >= 1:
            transmitter_params["stun-ip"] = stun_servers[0][0];
            transmitter_params["stun-port"] = stun_servers[0][1];

        if relay_info:
            transmitter_params["relay-info"] = relay_info

        self.fsstream.change_stream (transmitter, transmitter_params)

        self.fsstream.stream.newLocalCandidate.addCallback (False,
            self.new_candidate)

        self.obj.connect_to_signal ("SendersChanged",
            self.senders_changed,
            dbus_interface = CALL_STREAM)

        self.senders_changed()

    def new_candidate (self, candidate):
        # This should be (slightly) transport specific
        tpcandidate = (candidate.component_id, candidate.ip, candidate.port,
                {
                    "Type": candidate.type,
                    "Foundation": candidate.foundation,
                    "Protocol": candidate.proto,
                    "Priority": candidate.priority,
                    "Username": candidate.username,
                    "Password": candidate.password,
                    "BaseIP": candidate.base_ip,
                    "BasePort": candidate.base_port
                })
        self.obj.AddCandidates(dbus.Array([ tpcandidate ],
            signature='(usqa{sv})'),
            dbus_interface=CALL_STREAM_IFACE_MEDIA)

    def senders_changed (self, updates = None, removed = None):
        self_handle = self.conn.Get (CONN_INTERFACE, 'SelfHandle',
            dbus_interface=dbus.PROPERTIES_IFACE)
        senders = self.obj.Get (CALL_STREAM, 'Senders',
            dbus_interface=dbus.PROPERTIES_IFACE)
        direction = 0

        for k, v in senders.iteritems():
            if v == 2:
                bit = 1 if k == self_handle else 2
                direction |= bit

        #print '%s senders_changed: setting direction to %d' \
        #  % (self.fsstream.conference, direction)

        self.fsstream.conference.dump()

        self.fsstream.set_direction(direction)

        if self.senders_changed_cb != None:
            self.senders_changed_cb(self, senders, updates, removed)

    def set_sending (self, sending):
        self.obj.SetSending(sending, dbus_interface=CALL_STREAM)

    def set_senders_changed_callback (self, func):
        self.senders_changed_cb = func

    def set_src_pad_added_callback (self, func):
        self.fsstream.set_src_pad_added_callback (func)

    def destroy (self):
        self.fsstream.destroy()

class CallContent:
    def __init__ (self, bus, connection, object_path, conference, call):
        self.conn = connection
        self.bus = bus
        self.conference = conference
        self.call = call
        self.object_path = object_path
        self.obj = self.bus.get_object (self.conn.service_name, object_path)
        self.streams = []
        self.last_offer = None
        self.stream_added_cb = None
        self.stream_removed_cb = None
        # Mapping of farsight
        self.fsstreams = {}

        self.obj.connect_to_signal ('NewCodecOffer',
            self.codec_offer_cb,
            dbus_interface = CALL_CONTENT_IFACE_MEDIA)
        self.obj.connect_to_signal ("StreamAdded",
            self.add_stream, dbus_interface=CALL_CONTENT)
        self.obj.connect_to_signal ("StreamRemoved",
            self.remove_stream, dbus_interface=CALL_CONTENT)

        properties = self.obj.GetAll (CALL_CONTENT,
            dbus_interface=dbus.PROPERTIES_IFACE)

        self.mtype = mtype = properties["Type"]
        self.session = self.conference.create_session(mtype)
        self.session.get_codecs().addCallback (self.got_local_codecs)

        cstreams = properties["Streams"]
        for s in cstreams:
            self.add_stream(s)

        mproperties = self.obj.GetAll (CALL_CONTENT_IFACE_MEDIA,
            dbus_interface=dbus.PROPERTIES_IFACE)
        [ offer, codecs ] = mproperties["CodecOffer"]
        self.codec_offer_cb (offer, codecs)

    def add_stream(self, object_path):
        #print "Time to add a stream: %s" % object_path

        streamobj = self.bus.get_object (self.conn.service_name, object_path);

        properties = streamobj.GetAll (CALL_STREAM,
            dbus_interface = dbus.PROPERTIES_IFACE)

        senders = properties["Senders"]
        self_handle = self.conn.Get (CONN_INTERFACE, 'SelfHandle',
            dbus_interface=dbus.PROPERTIES_IFACE)

        for handle in senders.keys():
            if handle != self_handle:
                fsstream = self.ensure_fsstream (handle)
                cstream = CallStream (self.bus, self.conn, streamobj,
                        object_path, fsstream)

                self.streams.append (cstream)
                if self.stream_added_cb != None:
                    self.stream_added_cb (cstream, self)

    def remove_stream(self, object_path):
        for s in self.streams:
            if s.object_path == object_path:
                if self.stream_removed_cb != None:
                    self.stream_removed_cb (s, self)
                self.fsstreams.pop (s.fsstream.participant)
                s.destroy()
                self.streams.remove (s)

    def got_local_codecs (self, codecs):
        tpcodecs = []
        for c in codecs:
            params = {}
            for (key, value) in c.optional_params:
                params[key] = value

            tpcodecs.append (
              (c.id, c.encoding_name, c.clock_rate,
               c.channels, params))

        if self.last_offer is not None:
            offerobj = self.bus.get_object (self.conn.service_name,
                self.last_offer)
            self.last_offer = None

            offerobj.Accept(dbus.Array(tpcodecs, signature='(usuua{ss})'),
                dbus_interface=CALL_CONTENT_CODECOFFER)
        else:
            self.obj.SetCodecs(dbus.Array(tpcodecs, signature='(usuua{ss})'),
                dbus_interface=CALL_CONTENT_IFACE_MEDIA)

    def ensure_fsstream (self, handle):
        participant = self.call.ensure_participant (handle)
        if not participant in self.fsstreams:
            self.fsstreams[participant] = \
                FsStream (participant, self.conference, self.session)

        return self.fsstreams[participant]

    def codec_offer_cb(self, offer, codecs):
        if offer == "/":
            self.last_offer = None
            return

        self.last_offer = offer
        for handle, codecs in codecs.iteritems():
            fscodecs = []
            for c in codecs:
                fscodec = farsight.Codec (int(c[0]), str(c[1]),
                    self.session.type, int(c[2]))
                fscodec.channels = int(c[3])

                params = []
                for (key,value) in c[4].items():
                    params.append((str(key), str(value)))

                fscodec.optional_params = params
                fscodecs.append(fscodec)

            stream = self.ensure_fsstream (handle)
            stream.set_remote_codecs (fscodecs)

        self.session.get_codecs().addCallback (self.got_local_codecs)

class CallChannel:
    def __init__ (self, bus, connection, object_path, properties):
        self.conn = connection
        self.bus = bus
        self.conference = fs2.Conference ()
        self.contents = []
        self.call_state_changed_cb = None
        self.content_added_cb = None
        self.content_removed_cb = None
        self.members_updated_cb = None
        self.participants = {}
        self.members = {}

        self.obj = self.bus.get_object (self.conn.service_name, object_path)

        self.obj.connect_to_signal ("ContentAdded",
            self.content_added, dbus_interface=CHANNEL_TYPE_CALL)
        self.obj.connect_to_signal ("ContentRemoved",
            self.content_removed, dbus_interface=CHANNEL_TYPE_CALL)
        self.obj.connect_to_signal ("CallMembersChanged",
            self.call_members_changed, dbus_interface=CHANNEL_TYPE_CALL)

        properties = self.obj.GetAll (CHANNEL_TYPE_CALL,
            dbus_interface=dbus.PROPERTIES_IFACE)

        ccontents = properties["Contents"]

        self.state = properties["CallState"]
        self.flags = properties["CallFlags"]
        self.reason = properties["CallStateReason"]
        self.details = properties["CallStateDetails"]

        self.call_members_changed (properties["CallMembers"], [])

        self.obj.connect_to_signal ("CallStateChanged",
            self.state_changed_cb, dbus_interface=CHANNEL_TYPE_CALL)

        for c in ccontents:
            self.content_added(c)

        self.do_state_changes()

    def content_added(self, content, media_type = None):
        # Hack, old gabble had a bug where it would signal a content twice
        for c in self.contents:
            if c.object_path == content:
                return
        ccontent = CallContent (self.bus, self.conn, content,
            self.conference, self)
        self.contents.append (ccontent)

        if self.content_added_cb != None:
            self.content_added_cb(ccontent)

    def add_content(self, name, content_type):
        self.obj.AddContent(name, content_type,
            dbus_interface=CHANNEL_TYPE_CALL)

    def ringing(self):
        self.obj.Ringing(dbus_interface=CHANNEL_TYPE_CALL)

    def accept(self):
        self.obj.Accept(dbus_interface=CHANNEL_TYPE_CALL)

    def hangup(self, reason, hangup_reason, message):
        self.obj.Hangup (reason, hangup_reason, message,
            dbus_interface=CHANNEL_TYPE_CALL)

    def close(self):
        self.obj.Close(dbus_interface=CHANNEL_INTERFACE)


    def set_call_state_callback(self, func):
        self.call_state_changed_cb = func

    def do_state_changes (self):
        if self.state == CALL_STATE_ENDED and self.conference != None:
            self.conference.destroy()
            self.conference = None

        if (self.call_state_changed_cb != None):
            self.call_state_changed_cb(self)
        elif self.state == CALL_STATE_ENDED:
            self.close()

        print "* StateChanged:\n State: %s (%d)\n Flags: %s" % (
            call_state_to_s (self.state), self.state,
            call_flags_to_s (self.flags))

        print "\tReason: actor: %d reason: %d dbus_reason: '%s'" % (
            self.reason[0], self.reason[1], self.reason[2])

        print '\tDetails:'
        for key, value in self.details.iteritems():
            print "\t  %s: %s" % (key, value)
        else:
            print '\t  None'

    def state_changed_cb(self, state, flags, reason, details):
        self.state = state
        self.flags = flags
        self.reason = reason
        self.details = details

        self.do_state_changes()

    def content_removed(self, content):
        for c in self.contents:
            if c.object_path == content:
                content = c
                break
        else:
            DEBUG ("content_removed: Couldn't find content to remove")
            return

        self.contents.remove(content)

        if self.content_removed_cb != None:
            self.content_removed_cb(content)

        if self.conference != None:
            for s in content.streams:
                self.conference.remove_stream(s.stream)

        del content

    def ensure_participant (self, handle):
        if handle in self.participants:
            return self.participants[handle]

        contact = self.conn.InspectHandles (CONNECTION_HANDLE_TYPE_CONTACT,
            [handle])[0]
        participant = self.conference.create_participant(contact)
        self.participants[handle] = participant
        return participant

    def call_members_changed (self, members, removals):
        for x in removals:
            self.members.pop(x)
        for member, flags in members.iteritems():
            self.members[member] = flags
            self.ensure_participant (member)

        if self.members_updated_cb != None:
            self.members_updated_cb()
