# Farsight 2 glue

# Based on the farsight2 demo GUI program from the farsight2 source tree
# Copyright (C) 2007 Collabora, Nokia
# @author: Olivier Crete <olivier.crete@collabora.co.uk>
# @author: Sjoerd Simons <sjoerd.simons@collabora.co.uk>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

import sys, os, pwd, os.path
import socket
import threading
import weakref
import gc

from twisted.internet import defer
from twisted.words.xish.utility import CallbackList

try:
    import pygtk
    pygtk.require("2.0")

    import gtk, gtk.glade, gobject, gtk.gdk
    import gobject
except ImportError, e:
    raise SystemExit("PyGTK couldn't be found ! (%s)" % (e[0]))

try:
    import pygst
    pygst.require('0.10')

    import gst
except ImportError, e:
    raise SystemExit("Gst-Python couldn't be found! (%s)" % (e[0]))
try:
    import farsight
except ImportError, e:
    raise SystemExit("Farsight couldn't be found! (%s)" % (e[0]))


MYCNAME = "".join((pwd.getpwuid(os.getuid())[0],
                   "-" ,
                   str(os.getpid()),
                   "@",
                   socket.gethostname()))

CODECS = { farsight.MEDIA_TYPE_VIDEO: [ "X-VP8-GST", "THEORA", "H264", "H263",
            "H263-1998" ],
           farsight.MEDIA_TYPE_AUDIO: [ "SPEEX", "PCMA", "PCMU" ]
}

def make_video_sink(pipeline, xid, name, async=True):
    "Make a bin with a video sink in it, that will be displayed on xid."
    bin = gst.Bin("videosink_%d" % xid)
    sink = gst.element_factory_make("xvimagesink", name)
    sink.set_property("sync", async)
    sink.set_property("async", async)
    bin.add(sink)
    colorspace = gst.element_factory_make("ffmpegcolorspace")
    bin.add(colorspace)
    videoscale = gst.element_factory_make("videoscale")
    bin.add(videoscale)
    videoscale.link(colorspace)
    colorspace.link(sink)
    bin.add_pad(gst.GhostPad("sink", videoscale.get_pad("sink")))
    sink.set_data("xid", xid)
    return bin


class VideoSourceBin(gst.Bin):
    def __init__ (self, name = None):
        gst.Bin.__init__(self, name)

        if os.environ.has_key("VIDEO_DEVICE"):
            source = gst.element_factory_make("v4l2src")
            source.set_property("device", os.environ["VIDEO_DEVICE"])
        else:
            source = gst.element_factory_make("videotestsrc")
            source.set_property("is-live", True)

        self.add(source)

        colorspace = gst.element_factory_make("ffmpegcolorspace")
        self.add(colorspace)
        source.link(colorspace)

        rate = gst.element_factory_make("videorate", "sourcerate")
        self.add(rate)
        colorspace.link(rate)

        filter = gst.element_factory_make("capsfilter")
        filter.set_property("caps",
        gst.Caps("video/x-raw-yuv,width=[320,352],height=[240,288]," + \
            "framerate=(fraction)15/1,format=(fourcc)I420"))

        self.add(filter)
        rate.link(filter)

        tee = gst.element_factory_make("tee")
        self.tee = tee
        self.add(tee)

        filter.link(tee)

    def get_request_pad(self, name):
        pad = self.tee.get_request_pad (name)

        #queue = gst.element_factory_make("queue")
        #self.add(queue)
        #queue.sync_state_with_parent()
        #pad.link (queue.get_pad("sink"))

        #ghost = gst.GhostPad(pad.get_name(), queue.get_pad("src"))
        ghost = gst.GhostPad(pad.get_name(), pad)
        ghost.set_active (True)
        self.add_pad (ghost)
        return ghost


def make_audio_source():
    bin =gst.Bin()
    if os.environ.has_key("TESTAUDIO"):
        source = gst.element_factory_make("audiotestsrc")
        source.set_property("is-live", True)
    else:
        source = gst.element_factory_make("gconfaudiosrc")
    bin.add(source)
    pad = gst.GhostPad("src", source.get_pad("src"))
    bin.add_pad(pad)
    return bin

class ConferenceStream:
    def __init__(self, conference, session, fsstream):
        self.conference = conference
        self.session = session
        self._fsstream = fsstream
        self.newLocalCandidate = CallbackList()
        self.componentStateChanged = CallbackList()
        self.candidates = []
        self.sinks = []

        self.src_pad_hook = self.default_src_pad_added

        self._fsstream.connect("src-pad-added",
            self.src_pad_added)

    def new_local_candidate(self, candidate):
        self.candidates.append(candidate)
        self.newLocalCandidate.callback(candidate)

    def component_state_changed(self, component, state):
        self.componentStateChanged.callback(component, state)

    def add_remote_candidate(self, candidate):
        self._fsstream.set_remote_candidates([candidate])

    def set_remote_codecs(self, codecs):
        self._fsstream.set_remote_codecs(codecs)

    def default_src_pad_added (self, session, stream, pad, codec):
        if self.session.type == farsight.MEDIA_TYPE_VIDEO:
            sink = gst.element_factory_make("gconfvideosink")

        elif self.session.type == farsight.MEDIA_TYPE_AUDIO:
            sink = gst.element_factory_make("autoaudiosink")

        sinkpad = sink.get_pad ("sink")
        self.add_sink(sinkpad, sink)
        self.conference.pipeline.add( sink)
        sink.set_state(gst.STATE_PLAYING)
        pad.link(sinkpad)

    def src_pad_added(self, stream, pad, codec):
        print "+ Got media using: " + codec.to_string()
        self.src_pad_hook (self, stream, pad, codec)

    def set_direction(self, direction):
        self._fsstream.set_property("direction", direction)

    def add_sink(self, sinkpad, sink):
        self.sinks.append((sinkpad, sink))

    def destroy(self):
        for (sinkpad, sink) in self.sinks:
            peer = sinkpad.get_peer()
            if peer == None:
                continue

            peer.unlink (sinkpad)
            sink.set_locked_state(True)
            sink.set_state(gst.STATE_NULL)
            self.conference.pipeline.remove(sink)
        self.sinks = []

        # Ensure that there are no refs left to the _fsstream
        f = self._fsstream
        self._fsstream = None
        f.dispose()
        del f

        gc.collect()


class ConferenceSession:
    def __init__(self, conference, session, type):
        self.conference = conference
        self.type = type
        self.fssession = session
        self.codecs_ready_defereds = []
        self.source = None

        codecs = [
          farsight.Codec(farsight.CODEC_ID_ANY, name, type, 0)
            for name in CODECS[type]
        ]

        if type == farsight.MEDIA_TYPE_VIDEO:
           codecs.append(farsight.Codec(farsight.CODEC_ID_DISABLE,
            "DV", farsight.MEDIA_TYPE_VIDEO, 0))
           codecs.append(farsight.Codec(farsight.CODEC_ID_DISABLE,
            "MPV", farsight.MEDIA_TYPE_VIDEO, 0))
           codecs.append(farsight.Codec(farsight.CODEC_ID_DISABLE,
            "JPEG2000", farsight.MEDIA_TYPE_VIDEO, 0))
           codecs.append(farsight.Codec(farsight.CODEC_ID_DISABLE,
            "JPEG", farsight.MEDIA_TYPE_VIDEO, 0))
        elif type == farsight.MEDIA_TYPE_AUDIO:
            codecs.insert(0, farsight.Codec(farsight.CODEC_ID_ANY,
                "SPEEX:8000", farsight.MEDIA_TYPE_AUDIO, 8000))
            codecs.insert(0, farsight.Codec(farsight.CODEC_ID_ANY,
                "SPEEX", farsight.MEDIA_TYPE_AUDIO, 16000))

        # DV needs to be blacklisted for both audio and video
        codecs.append(farsight.Codec(farsight.CODEC_ID_DISABLE,
            "DV", type, 0))

        session.set_codec_preferences(codecs)

        if type == farsight.MEDIA_TYPE_VIDEO:
            source = VideoSourceBin()
            spad = source.get_request_pad("src%d")
        elif type == farsight.MEDIA_TYPE_AUDIO:
            source = make_audio_source()
            spad = source.get_pad("src")
        else:
            raise "Unknown source type"

        conference.pipeline.add(source)

        pad = session.get_property("sink-pad")

        spad.link(pad)

        source.set_state(gst.STATE_PLAYING)
        self.source = source

    def get_codecs(self):
        if self.fssession.get_property("codecs-ready"):
            return defer.succeed(self.fssession.get_property("codecs"))
        else:
            d = defer.Deferred()
            self.codecs_ready_defereds.append(d)
            return d

    def codecs_changed(self):
        id =  self.fssession.get_property("id")

        if self.fssession.get_property("codecs-ready"):
            defereds = self.codecs_ready_defereds
            self.codecs_ready_defereds = []

            codecs = self.fssession.get_property("codecs")
            map (lambda d: d.callback(codecs), defereds)

    def add_participant(self, participant, transmitter,
      transmitter_params = {}, direction = farsight.DIRECTION_BOTH):

        fsstream = self.fssession.new_stream (participant._fsparticipant,
            direction, transmitter, transmitter_params)
        stream = ConferenceStream(self.conference, self, fsstream)
        del fsstream

        self.conference.add_stream(stream)
        return stream

class ConferenceParticipant:
    def __init__(self, conference, fsparticipant, cname):
        self.conference = conference
        self._fsparticipant = fsparticipant
        self.cname = cname

class Conference:
    def __init__(self):
        self.pipeline = gst.Pipeline()

        self.notifier = notifier = farsight.ElementAddedNotifier()
        notifier.connect("element-added", self.element_added_cb)
        notifier.add(self.pipeline)

        self.pipeline.get_bus().add_watch(self.async_handler)
        self.conf = gst.element_factory_make("fsrtpconference")

        # Sets lets our own cname
        self.conf.set_property("sdes-cname", MYCNAME)
        self.pipeline.add(self.conf)
        self.pipeline.set_state(gst.STATE_PLAYING)

        self.fssessions = {}
        self.fsstreams = {}

    def add_stream(self, stream):
        self.fsstreams[stream._fsstream] = stream

    def remove_stream(self, stream):
        self.fsstreams.pop(stream._fsstream)
        stream.destroy()
        del stream

    def destroy(self):
        fsstreams = self.fsstreams.copy()
        for s in fsstreams.itervalues():
            self.remove_stream(s)
        self.pipeline.set_state(gst.STATE_NULL)

    def create_session(self, type):
        fs = self.conf.new_session(type)
        fs.set_property("no-rtcp-timeout", 0)
        s = ConferenceSession(self, fs, type)
        self.fssessions[fs] = s
        return s

    def create_participant(self, cname):
        fp = self.conf.new_participant(cname)
        p = ConferenceParticipant(self, fp, cname)
        return p

    def handle_element_farsight_error(self, message):
        print "***farsight error %d from %s: %s, %s" % \
            ( message.structure["error-no"],
              str(message.structure["src-object"]),
              message.structure["error-msg"],
              message.structure["debug-msg"])

    def handle_element_msg(self, message):
        dispatch = { }

        dispatch["farsight-codecs-changed"] = \
            lambda m : self.fssessions[m.structure["session"]].codecs_changed()

        dispatch["farsight-new-local-candidate"] = \
            lambda m : \
                self.fsstreams[m.structure["stream"]].new_local_candidate(
                    m.structure["candidate"])

        dispatch["farsight-component-state-changed"] = \
            lambda m : \
                self.fsstreams[m.structure["stream"]].component_state_changed(
                    m.structure["component"], m.structure["state"])

        dispatch["farsight-error"] = \
            lambda m : self.handle_element_farsight_error(m)

        name = message.structure.get_name()
        if dispatch.has_key(name):
            dispatch[name](message)
        else:
           #print "Unhandled element message %s: " % name
           pass

    def dump(self):
        gst.DEBUG_BIN_TO_DOT_FILE_WITH_TS(self.pipeline, 0xff,
            "fs2-muji-pipeline")

    def async_handler(self, bus, message):
        if message.type == gst.MESSAGE_ERROR:
            #print message.src.get_name(), ": ", message.parse_error()
            pass
        elif message.type == gst.MESSAGE_WARNING:
            #print message.src.get_name(), ": ", message.parse_warning()
            pass
        elif message.type == gst.MESSAGE_ELEMENT:
            self.handle_element_msg(message)
        return True

    def element_added_cb(self, notifier, bin, element):
        factory = element.get_factory()
        if factory == None:
            return

        name = factory.get_name()
        if name == "rtph264pay":
            element.set_property("config-interval", 5)
        elif name == "x264enc":
            element.set_property("byte-stream", True)
            element.set_property("bframes", 0)
            element.set_property("b-adapt", 0)
            element.set_property("cabac", 0)
            element.set_property("dct8x8", 0)
            element.set_property("bitrate", 128)
        elif name == "gstrtpbin":
            element.set_property("latency", 100)
        elif name == "theoraenc":
            element.set_property("bitrate", 128)
            element.set_property("rate-buffer", 1)
        elif name == "vp8enc":
            element.set_property("bitrate", 128000)
            element.set_property("speed", 2)
            element.set_property("max-latency", 1)
            element.set_property("error-resilient", True)

def get_available_transmitters():
    conf =  gst.element_factory_make("fsrtpconference")
    session = conf.new_session(farsight.MEDIA_TYPE_VIDEO)
    return session.list_transmitters()
